from patshared import PortType, TransportPosition


class PatchEngineOuter:
    @property
    def can_leave(self) -> bool:
        return False
    def write_existence_file(self):...
    def remove_existence_file(self):...
    def is_now_ready(self):...
    def associate_client_name_and_uuid(self, client_name: str, uuid: int):...
    def port_added(self, pname: str, ptype: PortType, pflags: int, puuid: int):...
    def port_renamed(self, ex_name: str, new_name: str, uuid=0):...
    def port_removed(self, port_name: str):...
    def jack_client_added(self, client_name: str):...
    def jack_client_removed(self, client_name: str):...
    def alsa_client_added(self, client_name: str):...
    def alsa_client_removed(self, client_name: str):...
    def metadata_updated(self, uuid: int, key: str, value: str):...
    def connection_added(self, connection: tuple[str, str]):...
    def connection_removed(self, connection: tuple[str, str]):...
    def server_stopped(self):...
    def send_transport_position(self, tpos: TransportPosition):...
    def send_dsp_load(self, dsp_load: int):...
    def send_one_xrun(self):...
    def send_buffersize(self, buffer_size: int):...
    def send_samplerate(self, samplerate: int):...
    def send_pretty_names_locked(self, locked: bool):...
    def send_server_lose(self):...
    def server_restarted(self):...
    def make_one_shot_act(self, one_shot_act: str):...